# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
# Author: Thomi Richards
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

from testscenarios import TestWithScenarios
from testtools import TestCase
from testtools.matchers import Contains, Equals, raises

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from windowmocker.plugins import (
    base,
    get_plugin_by_name,
    register_plugin,
    register_named_plugin,
    unregister_plugin_by_name
    )

class DummyAppPlugin(base.ApplicationTypePluginBase):

    Name = "Dummy"

    def create_application(self):
        pass

    def create_window(self, window_spec):
        pass


class PluginRegistryTests(TestCase):

    def register_plugin(self, plugin_class):
        register_plugin(plugin_class)
        self.addCleanup(unregister_plugin_by_name, plugin_class.Name)

    def register_named_plugin(self, plugin_name, plugin_class):
        register_named_plugin(plugin_name, plugin_class)
        self.addCleanup(unregister_plugin_by_name, plugin_name)

    def test_returns_None_on_unknown_plugin_name(self):
        plugin = get_plugin_by_name("unknown")
        self.assertThat(plugin, Equals(None))

    def test_can_register_plugin(self):
        self.register_plugin(DummyAppPlugin)
        plugin = get_plugin_by_name("Dummy")

        self.assertThat(plugin, Equals(DummyAppPlugin))

    def test_can_register_named_plugin(self):
        self.register_named_plugin("Dummy", DummyAppPlugin)
        plugin = get_plugin_by_name("Dummy")

        self.assertThat(plugin, Equals(DummyAppPlugin))

    def test_can_create_empty_plugin(self):
        # must not raise NotImplementedError:
        DummyAppPlugin({})

    def test_base_raises_NotImplemented(self):
        fn = lambda: base.ApplicationTypePluginBase({"Windows": {}})

        self.assertThat(fn, raises(NotImplementedError))

    @patch.object(DummyAppPlugin, "create_application")
    def test_calls_create_application(self, mock_create_app):
        DummyAppPlugin({})

        mock_create_app.assert_called_once_with()

    @patch.object(DummyAppPlugin, "create_window")
    def test_window_created_by_default(self, mock_create_window):
        DummyAppPlugin({})

        self.assertThat(mock_create_window.call_count, Equals(1))
        self.assertThat(mock_create_window.call_args[0][0],
                        Equals(DummyAppPlugin._window_default))

    @patch.object(DummyAppPlugin, "create_window")
    def test_calls_create_window_with_defaults_added(self, mock_create_window):
        DummyAppPlugin({})

        self.assertThat(mock_create_window.call_count, Equals(1))
        self.assertThat(mock_create_window.call_args[0][0],
                        Equals(DummyAppPlugin._window_default))

    @patch.object(DummyAppPlugin, "create_window")
    def test_passes_params_to_create_window(self, mock_create_window):
        DummyAppPlugin({"Title": "Hello World"})

        self.assertThat(mock_create_window.call_count, Equals(1))
        passed_spec = mock_create_window.call_args[0][0]

        self.assertThat(passed_spec, Contains("Title"))
        self.assertThat(passed_spec["Title"], Equals("Hello World"))

    @patch.object(DummyAppPlugin, "create_window")
    def test_can_create_many_windows(self, mock_create_window):
        DummyAppPlugin([
                {"Title": "Window 1"},
                {"Title": "Window 2"},
                ])

        self.assertThat(mock_create_window.call_count, Equals(2))
        self.assertThat(mock_create_window.call_args_list[0][0][0]["Title"],
                        Equals("Window 1"))
        self.assertThat(mock_create_window.call_args_list[1][0][0]["Title"],
                        Equals("Window 2"))


class PluginWindowDefaultParameterTests(TestWithScenarios, TestCase):

    """Tests for various default values for required parameters."""

    scenarios = [
        ("Window Title",
            {'name': 'Title', 'expected': 'Default Window Title'}),
        ("Maximised", {'name': 'Maximized', 'expected': False}),
        ("Minimized", {'name': 'Minimized', 'expected': False}),
        ("MinimizeImmediatelyAfterShow",
            {'name': 'MinimizeImmediatelyAfterShow', 'expected': False}),
        ("Menus", {'name': 'Menu', 'expected': []}),
        ]

    @patch.object(DummyAppPlugin, "create_window")
    def test_default_parameter(self, mock_create_window):
        DummyAppPlugin({"Windows": {}})
        window_spec = mock_create_window.call_args[0][0]

        self.assertThat(window_spec, Contains(self.name))
        self.assertThat(window_spec[self.name], Equals(self.expected))



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
# Author: Thomi Richards
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

import logging
import sys


logging.basicConfig()
logger = logging.getLogger(__name__)

def create_qt4_plugin(data):
    from windowmocker.plugins.qt4 import QtPlugin as Qt4Plugin
    return Qt4Plugin(data)

def create_qt5_plugin(data):
    from windowmocker.plugins.qt5 import QtPlugin as Qt5Plugin
    return Qt5Plugin(data)

_plugin_registry = {
    'Qt4': create_qt4_plugin,
    'Qt5': create_qt5_plugin,
}

_default_plugin = create_qt4_plugin

def get_plugin_by_name(name):
    """Return the application type plugin class by name."""
    global _plugin_registry
    plugin = _plugin_registry.get(name, None)
    if not plugin:
        logger.error("No plugin was found with name '%s'", name)
    return plugin

def get_default_plugin_name():
    return 'Qt4'

def get_registered_plugin_names():
    """Get a list of registered plugin names."""
    global _plugin_registry
    if sys.version_info[0] < 3:
        del _plugin_registry['Qt5']
    return _plugin_registry.keys()


def register_named_plugin(plugin_name, plugin_class):
    global _plugin_registry

    old_plugin = _plugin_registry.get(plugin_name, None)
    if old_plugin:
        logger.warning("Plugin %r is already registered with name '%s'. Overwriting it with %r.",
            old_plugin, plugin_name, plugin_class)

    _plugin_registry[plugin_name] = plugin_class

def register_plugin(plugin_class):
    """Register plugin_class. Probably not ever needed, except by tests.

    This function is deprecated and may be removed in a future release.
    Use `register_named_plugin` instead."""
    global _plugin_registry

    old_plugin = _plugin_registry.get(plugin_class.Name, None)
    if old_plugin:
        logger.warning("Plugin %r is already registered with name '%s'. Overwriting it with %r.",
            old_plugin, plugin_class.Name, plugin_class)

    _plugin_registry[plugin_class.Name] = plugin_class

def unregister_plugin_by_name(name):
    """Unregister a plugin with a given name. Probably not ever needed, except by tests."""
    global _plugin_registry

    if name not in _plugin_registry:
        logger.error("Cannot unregister plugin '%s' since it's not already registered.",
            name)
    else:
        _plugin_registry.pop(name)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

from __future__ import print_function
import logging
from PyQt5 import QtCore, QtWidgets
import six
import sys

from windowmocker.plugins.base import ApplicationTypePluginBase

logger = logging.getLogger(__name__)

class QtPlugin(ApplicationTypePluginBase):

    Name = "Qt5"

    def create_application(self):
        args = sys.argv[:1]
        if "-testability" in sys.argv:
            args.append("-testability")
        self.app = QtWidgets.QApplication(args)

    def create_window(self, window_spec):
        win = QtWidgets.QMainWindow()
        win.setWindowTitle(window_spec["Title"])

        if 'Menu' in window_spec:
            menu_bar = win.menuBar()
            self._create_menus(menu_bar, window_spec['Menu'])

        control_spec = window_spec.get('Contents', None)
        if control_spec is not None:
            controls = self._create_controls(control_spec)
            win.setCentralWidget(controls)

        if window_spec['Maximized']:
            win.showMaximized()
        if window_spec['Minimized']:
            win.showMinimized()

        if not window_spec['Maximized'] and not window_spec['Minimized']:
            win.showNormal()

        # In some cases, it's also needed to have the application start
        # normally but minimize itself right after starting
        if window_spec['MinimizeImmediatelyAfterShow']:
            win.setWindowState(QtCore.Qt.WindowMinimized)

        if not getattr(self, 'windows', None):
            self.windows = []
        self.windows.append(win)

    def _create_menus(self, parent, menu_spec):
        for item in menu_spec:
            if type(item) is dict:
                title = item.get('Title', 'Default Title')
                menu = parent.addMenu(title)
                if 'Menu' in item:
                    self._create_menus(menu, item['Menu'])
            elif isinstance(item, six.string_types):
                parent.addAction(item)
            else:
                logger.error("Invalid menu item - is not a dict or string: %r", item)

    def _create_controls(self, control_spec):
        if control_spec == 'TextEdit':
            return QtWidgets.QTextEdit()
        elif control_spec == 'MouseTest':
            return MouseTestWidget()



    def run(self):
        self.app.exec_()


class MouseTestWidget(QtWidgets.QWidget):

    def __init__(self):
        super(MouseTestWidget, self).__init__()
        self.setAttribute(QtCore.Qt.WA_AcceptTouchEvents)
        self.button_control = QtWidgets.QLabel(self)
        self.button_control.setAttribute(QtCore.Qt.WA_AcceptTouchEvents)
        self.button_control.setObjectName("button_status")
        self.position_control = QtWidgets.QLabel(self)
        self.position_control.setAttribute(QtCore.Qt.WA_AcceptTouchEvents)
        self.position_control.setObjectName("position_status")
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.button_control)
        layout.addWidget(self.position_control)

        self.setMouseTracking(True)
        self.button_control.setMouseTracking(True)
        self.position_control.setMouseTracking(True)

    def event(self, event):
        if event.type() == QtCore.QEvent.TouchBegin:
            self.touchBegin(event)
        elif event.type() == QtCore.QEvent.TouchUpdate:
            self.touchUpdate(event)
        elif event.type() == QtCore.QEvent.TouchEnd:
            self.touchEnd(event)
        else:
            return super(MouseTestWidget, self).event(event)
        return True

    def touchBegin(self, event):
        self.button_control.setText("Touch Press")

    def touchUpdate(self, event):
        pass

    def touchEnd(self, event):
        self.button_control.setText("Touch Release")

    def mousePressEvent(self, mouseEvent):
        btn_name = self._get_button_string(mouseEvent.button())
        self.button_control.setText("Press " + btn_name)

    def mouseReleaseEvent(self, mouseEvent):
        btn_name = self._get_button_string(mouseEvent.button())
        self.button_control.setText("Release " + btn_name)

    def mouseMoveEvent(self, mouseEvent):
        pos = mouseEvent.globalPos()
        self.position_control.setText("Position: ({}, {})".format(pos.x(), pos.y()))

    def wheelEvent(self, wheelEvent):
        direction = "down" if wheelEvent.delta() < 0 else "up"
        self.button_control.setText("Wheel {}".format(direction))

    def _get_button_string(self, btn):
        if btn == QtCore.Qt.LeftButton:
            return "Left"
        elif btn == QtCore.Qt.RightButton:
            return "Right"
        elif btn == QtCore.Qt.MiddleButton:
            return "Middle"
        else:
            return "Other"



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
# Author: Thomi Richards
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

"""Functional tests for the window-mocker application."""

import json
import os
import subprocess
import tempfile

from time import sleep

from testtools import TestCase
from testtools.matchers import Contains, Equals, MatchesAny


def get_app_path():
    """Return the full path to the test application.

    This function prefers the local copy if these tests are being run from
    source.

    """
    source_path = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..',
            'bin',
            'window-mocker'))
    if os.path.exists(source_path):
        return source_path

    return subprocess.check_output(['which', 'window-mocker'],
                                   universal_newlines=True)


def run_app_with_args(*args):
    os.putenv('PYTHONPATH',
        os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                '..')
            )
        )
    application = [get_app_path()]
    application.extend(args)
    process = subprocess.Popen(
            application,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
            universal_newlines=True,
            )

    stdout = stderr = ""
    for i in range(3):
        retcode = process.poll()
        if retcode is not None:
            break
        if i == 1:
            print("Terminating process.")
            process.terminate()
        if i == 2:
            print("Killing process.")
            process.kill()
        sleep(1)
    stdout, stderr = process.communicate()
    return (retcode, stdout, stderr)


class FunctionalTests(TestCase):

    """Functional tests for the window-mocker application."""

    def run_with_spec_content(self, content_string):
        fd, temp_filename = tempfile.mkstemp()
        self.addCleanup(os.remove, temp_filename)
        os.close(fd)
        with open(temp_filename, 'w') as f:
            f.write(content_string)
        return run_app_with_args(temp_filename)

    def test_spec_with_menus(self):
        window_spec = {
            "Title": "Test menu application",
            "Menu": ["Search entry", "Quit"],
        }
        fd, temp_filename = tempfile.mkstemp()
        self.addCleanup(os.remove, temp_filename)
        os.close(fd)
        with open(temp_filename, 'w') as f:
            f.write(json.dumps(window_spec))
        return run_app_with_args(temp_filename)

    def test_errors_with_missing_file(self):
        retcode, stdout, stderr = run_app_with_args("bad_file_name")
        self.assertThat(stdout, Equals(''))
        self.assertThat(stderr, Equals('Error: Unable to open file: No such file or directory\n'))
        self.assertThat(retcode, Equals(1))

    def test_errors_with_inaccessible_file(self):
        fd, temp_filename = tempfile.mkstemp()
        self.addCleanup(os.remove, temp_filename)
        os.close(fd)
        with open(temp_filename, 'w'):
            pass
        os.chmod(temp_filename, 0)

        retcode, stdout, stderr = run_app_with_args(temp_filename)
        self.assertThat(stdout, Equals(''))
        self.assertThat(stderr, Equals('Error: Unable to open file: Permission denied\n'))
        self.assertThat(retcode, Equals(1))

    def test_errors_on_empty_file(self):
        retcode, stdout, stderr = self.run_with_spec_content("")

        self.assertThat(stdout, Equals(''))
        self.assertThat(retcode, Equals(3))
        self.assertThat(
            stderr,
            MatchesAny(
                Equals('Error: Unable to process request: No JSON object could be decoded\n'),
                Equals('Error: Unable to process request: Expecting value: line 1 column 1 (char 0)\n'),
            )
        )

    def test_accepts_testability_command_line_argument(self):
        retcode, stdout, stderr = run_app_with_args("--help")
        self.assertThat(stdout, Contains("-testability"))

    def test_accepts_plugin_type_command_line_argument(self):
        retcode, stdout, stderr = run_app_with_args("--help")
        self.assertThat(stdout, Contains("Qt4"))



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
# Author: Thomi Richards
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

"""Base classes for application type implementations."""

import logging


logger = logging.getLogger(__name__)


class ApplicationTypePluginBase(object):

    """A base class for all application type plugins."""

    _window_default = {
        'Title': "Default Window Title",
        'Maximized': False,
        'Minimized': False,                    # Start application minimized
        'MinimizeImmediatelyAfterShow': False, # After starting, minimize immediately
        'Menu': [],
        'Contents': None
    }

    def __init__(self, spec_dictionary):
        self.create_application()

        if not spec_dictionary:
            logger.warning("Specification contains no window specification. Creating a default window.")

        if isinstance(spec_dictionary, dict):
            window_spec = self._create_window_spec(spec_dictionary)
            self.create_window(window_spec)
        elif isinstance(spec_dictionary, list):
            for window in spec_dictionary:
                window_spec = self._create_window_spec(window)
                self.create_window(window_spec)

    def _create_window_spec(self, window_params):
        """Return a dictionary that contains window_params with all the defaults
        added.

        """
        defaults = self._window_default.copy()
        defaults.update(window_params)
        return defaults

    def create_application(self):
        """Create an application."""
        raise NotImplementedError(
            "This method must be implemented by child classes.")

    def create_window(self, window_spec):
        """Create a window, according to 'window_spec'."""
        raise NotImplementedError(
            "This method must be implemented by child classes.")

    def run(self):
        """Enter the event loop - run the application & windows."""
        raise NotImplementedError(
            "This method must be implemented by child classes.")



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
# Author: Thomi Richards
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.


import os
import tempfile

from io import StringIO

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from testtools import TestCase
from testtools.matchers import raises

import windowmocker


class TestFilePathLoading(TestCase):

    """Tests the ability to create apps from a file path."""

    def test_create_application_from_path_raises_IOError_on_missing_file(self):
        """create_application_from_path function must raise IOError when asked
        to open a file that does not exist.
        """
        self.assertThat(lambda: windowmocker.create_application_from_path("_does_not_exist"), raises(IOError))

    def test_create_application_from_path_raises_PermError_on_bad_perms(self):
        """create_application_from_path function must raise PermissionError
        when the file's permissions do not allow it to be read.
        """
        fd, temp_filename = tempfile.mkstemp()
        self.addCleanup(os.remove, temp_filename)
        os.close(fd)
        with open(temp_filename, 'w'):
            pass
        os.chmod(temp_filename, 0)

        self.assertThat(
            lambda: windowmocker.create_application_from_path(temp_filename),
            raises(IOError))


class TestFileObjectLoading(TestCase):

    """Tests the ability to create apps from a file-like object."""

    def test_raises_attributeerror_on_bad_parameter_type(self):
        """create_application_from_file must raise a AttributeError if we give it
        something other than a file-like object.

        """
        self.assertThat(lambda: windowmocker.create_application_from_file("string"),
            raises(AttributeError))
        self.assertThat(lambda: windowmocker.create_application_from_file(123),
            raises(AttributeError))
        self.assertThat(lambda: windowmocker.create_application_from_file(123.456),
            raises(AttributeError))
        self.assertThat(lambda: windowmocker.create_application_from_file(True),
            raises(AttributeError))

    @patch('windowmocker.create_application_from_data')
    def test_can_load_ascii_json(self, load_mock):
        windowmocker.create_application_from_file(StringIO(u'["Hello", "World"]'))
        load_mock.assert_called_once_with(["Hello", "World"], None)

    @patch('windowmocker.create_application_from_data')
    def test_can_load_unicode_json(self, load_mock):
        windowmocker.create_application_from_file(
            StringIO(u'["ɥǝןןo", "ʍoɹןp"]'))
        load_mock.assert_called_once_with(
            [u'\u0265\u01dd\u05df\u05dfo', u'\u028do\u0279\u05dfp'], None)

    def test_raises_runtime_on_empty_file(self):
        # BAW 2014-01-31: We cannot assert on the exact error message
        # returned, because it's different between Python 3.3 and 3.4.
        # http://bugs.python.org/issue20453
        self.assertThat(
            lambda: windowmocker.create_application_from_file(StringIO()),
            raises(RuntimeError))



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
# Author: Thomi Richards
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.


import json
import logging
import os
import subprocess
import tempfile
from threading import Thread


from windowmocker import plugins

__all__ = (
    'create_application_from_data',
    'create_application_from_file',
    'create_application_from_path',
    'launch_window_mocker_with_data',
    )


logger = logging.getLogger(__name__)


def create_application_from_path(path, app_type_name=None):
    """Create and return an Application instance, reading settings from path.

    app_type_name is the plugin name for the type of application you want to
    create. If set to None (the default), the default plugin will be used.

    This function raises an IOError if the specified path does not exist, or is
    not readable.

    """
    with open(path, 'r') as fp:
        return create_application_from_file(fp, app_type_name)


def create_application_from_file(file_object, app_type_name=None):
    """Create and return an Application instance, reading settings from the
    file-like object 'file_object'.

    app_type_name is the plugin name for the type of application you want to
    create. If set to None (the default), the default plugin will be used.

    This function raises AttributeError if the object does not have a 'read()'
    method.

    """
    try:
        data = json.load(file_object)
    except ValueError as e:
        raise RuntimeError(str(e))
    return create_application_from_data(data, app_type_name)


def create_application_from_data(data=None, app_type_name=None):
    """Create an application object from 'data'.

    If data is None (the default), a default application is constructed.

    app_type_name is the plugin name for the type of application you want to
    create. If set to None (the default), the default plugin will be used.

    This function raises RuntimeErrors if the data does not conform to the
    expected format.

    """
    if data is None:
        data = {}
    if type(data) not in (dict, list):
        raise RuntimeError(
            "Top-level data structure must be a dictionary or list, not %r" %
            (type(data).__name__))

    if app_type_name is None:
        app_type_name = plugins.get_default_plugin_name()

    app_plugin = plugins.get_plugin_by_name(app_type_name)
    if app_plugin is None:
        raise RuntimeError("Type name '%s' is invalid" % app_type_name)

    app_instance = app_plugin(data)
    app_instance.run()


def launch_window_mocker_with_data(data=None):
    """Run window-mocker with the contents of 'data' as the window spec file.

    If data is None (the default), a default window will be created. This
    function returns a threading.Thread object that has been started. While the
    window is active, the thread will remain running.

    This function writes 'data' to a temporary file on disk, and calls
    window-mocker on that spec file. The file is cleaned up afterwards.

    window-mocker must be in $PATH.

    """
    if data is None:
        data = {}

    if type(data) not in (dict, list):
        raise RuntimeError(
            "Top-level data structure must be a dictionary or list, not %r" %
            (type(data).__name__))

    def run_thread():
        fd, temp_filename = tempfile.mkstemp()
        os.close(fd)
        try:
            with open(temp_filename, 'w') as f:
                json.dump(data, f)
            subprocess.check_call(["window-mocker", temp_filename])
        finally:
            os.remove(temp_filename)

    thread = Thread(target=run_thread)
    thread.start()
    return thread



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012-2014 Canonical, Ltd.
# Author: Thomi Richards
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

import DistUtilsExtra.auto
from setuptools import find_packages

DistUtilsExtra.auto.setup(
    name='window-mocker',
    version='1.4',
    description=open('README').read(),
    author='Thomi Richards',
    author_email='thomi.richards@canonical.com',
    license='GPLv3',
    packages=find_packages(exclude=('tests',)),
    test_suite='tests',
)
